/*
 * Copyright 2024 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef BAIDUSPEECHENGINE_P_H
#define BAIDUSPEECHENGINE_P_H

#include <thread>
#include <curl/curl.h>
#include <kylin-ai/ai-engine/cloudspeechengine.h>

#include "cpr/cpr.h"

// 开始：ReconnectSettings
class ReconnectSettings {
public:
    using Callback = std::function<void()>;

public:
    void setCallback(const Callback &callback) { callback_ = callback; }

    void setLimitRetryCount(int32_t count) { limitRetryCount_ = count; }

    void setRetryIntervalMs(int32_t ms) { retryIntervalMs_ = ms; }

    /**
     * 超过重试次数，不重试
     * 第一次运行，不重试
     */
    void retryOnce();

    void resetRetryParam();

    bool canRetry() const;

    bool isRetrying() const { return retrying_; }

private:
    void callCallback() const;

    void initRetryParam();

private:
    Callback callback_;
    int32_t limitRetryCount_{5}; // 限制重试次数
    int32_t retryIntervalMs_{1000};
    int64_t currentTimestamp_{0};
    std::atomic<int32_t> currentRetryCount_{0}; // 当前重试次数
    std::atomic<bool> retrying_{false};
};
// 结束：ReconnectSettings

class BaiduSpeechEnginePrivate
    : public ai_engine::speech::AbstractCloudSpeechEngine {
public:
    BaiduSpeechEnginePrivate();
    ~BaiduSpeechEnginePrivate() override;
    std::string engineName() const override;
    std::string currentModelName() const override;
    int maxConcurrentOnceRecognitionTasks() const override;
    int maxConcurrentContinuousRecognitionTasks() const override;
    int maxConcurrentOnceSynthesisTasks() const override;
    int maxConcurrentContinuousSynthesisTasks() const override;

    void setConfig(const std::string &config) override;
    std::string modelInfo() const override;

    void setRecognizingCallback(
        ai_engine::speech::RecognitionResultCallback callback) override;
    void setRecognizedCallback(
        ai_engine::speech::RecognitionResultCallback callback) override;

    void setSynthesizingCallback(
        ai_engine::speech::SynthesisResultCallback callback) override;
    void setSynthesizedCallback(
        ai_engine::speech::FinishedCallback callback) override;

    // 流式语音识别
    bool
    initContinuousRecognitionModule(ai_engine::EngineError &error) override;
    bool startContinuousRecognition(const std::string &params,
                                    ai_engine::EngineError &error) override;
    bool
    writeContinuousRecognitionAudioData(const std::vector<char> &audioData,
                                        ai_engine::EngineError &error) override;
    bool stopContinuousRecognition(ai_engine::EngineError &error) override;
    bool
    destroyContinuousRecognitionModule(ai_engine::EngineError &error) override;

    // 一次语音识别
    bool initRecognizeOnceModule(ai_engine::EngineError &error) override;
    bool recognizeOnce(const std::string &params,
                       const std::vector<char> &audioData,
                       ai_engine::EngineError &error) override;
    bool destroyRecognizeOnceModule(ai_engine::EngineError &error) override;

    // 流式语音合成
    bool initContinuousSynthesisModule(ai_engine::EngineError &error) override;
    bool startContinuousSynthesis(const std::string &params,
                                  ai_engine::EngineError &error) override;
    bool writeContinuousSynthesisText(const std::string &text,
                                      ai_engine::EngineError &error) override;
    bool stopContinuousSynthesis(ai_engine::EngineError &error) override;
    bool
    destroyContinuousSynthesisModule(ai_engine::EngineError &error) override;

    // 一次语音合成
    bool initSynthesizeOnceModule(ai_engine::EngineError &error) override;
    bool synthesizeOnce(const std::string &params, const std::string &text,
                        ai_engine::EngineError &error) override;
    bool destroySynthesizeOnceModule(ai_engine::EngineError &error) override;

private:
    // 回调函数相关
    void callRecognizingCallback(
        const ai_engine::speech::RecognitionResult &result) const;

    void callRecognizedCallback(
        const ai_engine::speech::RecognitionResult &result) const;

    void callSynthesizingCallback(
        const ai_engine::speech::SynthesisResult &result) const;

    void callSynthesizedCallback(ai_engine::speech::ResultType type) const;

private:
    enum class OperationType {
        RecognizeOnce,
        ContinuousRecognition,
        SynthesizeOnce,
        ContinuousSynthesis
    };

    std::string &getAccessToken();
    bool generateAccessToken();

    bool doSendWebSocketData(const char *data, int dataLength, int type);

    bool doSendEndFlag();

    void setRecognitionParams(const std::string &params);
    bool writeOnceRecognitionResult(const std::string &data);
    bool doReceiveContinuousRecognitionResultData();
    bool sendAuthParamsData();

    bool connectWebSocket();
    void onWebSocket();
    bool testWebSocket();
    void closeWebsocket();

    void setSynthesisParams(const std::string &params);
    bool doRequestSynthesize(const std::string &text,
                             ai_engine::EngineError &error, OperationType type);
    bool writeSynthesisResult(const std::string &data, OperationType type);

    void releaseResources();
    bool processCprResponse(cpr::Response &response, OperationType type);

    bool currentStatus();
    void runCallbackWithError(const ai_engine::EngineError &error,
                              OperationType type);
    // 流式语音识别的重连
    void continuousAsrReconnectWebSockets();

private:
    bool needRefreshAccessToken_{false};

    std::string appId_;
    std::string apiKey_;
    std::string secretKey_;
    std::string accessToken_;

    ai_engine::speech::RecognitionResultCallback recognizingCallback_;
    ai_engine::speech::RecognitionResultCallback recognizedCallback_;
    ai_engine::speech::SynthesisResultCallback synthesizingCallback_;
    ai_engine::speech::FinishedCallback synthesizedCallback_;

    std::string format_;
    int rate_{0};
    int channel_{0};

    const int maxWebDelay_{1000 * 10};
    int currentWebDelay_{0};
    bool isPinging_{false};

    int speed_{5};
    int volume_{5};
    int pitch_{5};

    std::atomic<bool> isStopped_{false};

    CURL *currentCurl_{nullptr};
    std::thread *currentThread_{nullptr};
    ai_engine::EngineError currentError_;

    std::mutex mutex_;

    ReconnectSettings reconnectSettings_;

};

#endif
